"""frosted/test/test_api.py.

Tests all major functionality of the Frosted API

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""
from __future__ import absolute_import, division, print_function, unicode_literals

import os
import sys
import tempfile
from io import StringIO

import pytest
from frosted.api import check_path, check_recursive
from frosted.messages import UnusedImport
from frosted.reporter import Reporter
from pies.overrides import *

from .utils import LoggingReporter, Node


def test_syntax_error():
    """syntax_error reports that there was a syntax error in the source file.

    It reports to the error stream and includes the filename, line number, error message, actual line of source and a
    caret pointing to where the error is

    """
    err = StringIO()
    reporter = Reporter(None, err)
    reporter.syntax_error('foo.py', 'a problem', 3, 7, 'bad line of source')
    assert ("foo.py:3:7: a problem\n"
            "bad line of source\n"
            "       ^\n") == err.getvalue()


def test_syntax_errorNoOffset():
    """syntax_error doesn't include a caret pointing to the error if offset is passed as None."""
    err = StringIO()
    reporter = Reporter(None, err)
    reporter.syntax_error('foo.py', 'a problem', 3, None, 'bad line of source')
    assert ("foo.py:3: a problem\n"
            "bad line of source\n") == err.getvalue()


def test_multiLineSyntaxError():
    """ If there's a multi-line syntax error, then we only report the last line.

    The offset is adjusted so that it is relative to the start of the last line

    """
    err = StringIO()
    lines = ['bad line of source', 'more bad lines of source']
    reporter = Reporter(None, err)
    reporter.syntax_error('foo.py', 'a problem', 3, len(lines[0]) + 7,
                          '\n'.join(lines))
    assert ("foo.py:3:6: a problem\n" +
            lines[-1] + "\n" +
            "      ^\n") == err.getvalue()


def test_unexpected_error():
    """unexpected_error reports an error processing a source file."""
    err = StringIO()
    reporter = Reporter(None, err)
    reporter.unexpected_error('source.py', 'error message')
    assert 'source.py: error message\n' == err.getvalue()


def test_flake():
    """flake reports a code warning from Frosted.

    It is exactly the str() of a frosted.messages.Message

    """
    out = StringIO()
    reporter = Reporter(out, None)
    message = UnusedImport('foo.py', Node(42), 'bar')
    reporter.flake(message)
    assert out.getvalue() == "%s\n" % (message,)


def make_temp_file(content):
    """Make a temporary file containing C{content} and return a path to it."""
    _, fpath = tempfile.mkstemp()
    if not hasattr(content, 'decode'):
        content = content.encode('ascii')
    fd = open(fpath, 'wb')
    fd.write(content)
    fd.close()
    return fpath


def assert_contains_errors(path, errorList):
    """Assert that provided causes at minimal the errors provided in the error list."""
    error = StringIO()
    count = check_path(path, Reporter(sys.stdout, error))
    error_string = error.getvalue()
    assert len(errorList) >= count
    for error in errorList:
        assert error in error_string


def get_errors(path):
    """Get any warnings or errors reported by frosted for the file at path."""
    log = []
    reporter = LoggingReporter(log)
    count = check_path(path, reporter)
    return count, log


def test_missingTrailingNewline():
    """Source which doesn't end with a newline shouldn't cause any exception to

    be raised nor an error indicator to be returned by check.

    """
    fName = make_temp_file("def foo():\n\tpass\n\t")
    assert_contains_errors(fName, [])


def test_check_pathNonExisting():
    """check_path handles non-existing files"""
    count, errors = get_errors('extremo')
    assert count == 1
    assert errors == [('unexpected_error', 'extremo', 'No such file or directory')]


def test_multilineSyntaxError():
    """Source which includes a syntax error which results in the raised SyntaxError.

    text containing multiple lines of source are reported with only
    the last line of that source.

    """
    source = """\
def foo():
    '''

def bar():
    pass

def baz():
    '''quux'''
"""
    # Sanity check - SyntaxError.text should be multiple lines, if it
    # isn't, something this test was unprepared for has happened.
    def evaluate(source):
        exec(source)
    try:
        evaluate(source)
    except SyntaxError:
        e = sys.exc_info()[1]
        assert e.text.count('\n') > 1
    else:
        assert False

    sourcePath = make_temp_file(source)
    assert_contains_errors(
        sourcePath,
        ["""\
%s:8:10: invalid syntax
    '''quux'''
          ^
""" % (sourcePath,)])


def test_eofSyntaxError():
    """The error reported for source files which end prematurely causing a
    syntax error reflects the cause for the syntax error.

    """
    sourcePath = make_temp_file("def foo(")
    assert_contains_errors(sourcePath, ["""\
%s:1:8: unexpected EOF while parsing
def foo(
        ^
""" % (sourcePath,)])


def test_nonDefaultFollowsDefaultSyntaxError():
    """ Source which has a non-default argument following a default argument

    should include the line number of the syntax error
    However these exceptions do not include an offset

    """
    source = """\
def foo(bar=baz, bax):
    pass
"""
    sourcePath = make_temp_file(source)
    last_line = '       ^\n' if sys.version_info >= (3, 2) else ''
    column = '7:' if sys.version_info >= (3, 2) else ''
    assert_contains_errors(sourcePath, ["""\
%s:1:%s non-default argument follows default argument
def foo(bar=baz, bax):
%s""" % (sourcePath, column, last_line)])


def test_nonKeywordAfterKeywordSyntaxError():
    """Source which has a non-keyword argument after a keyword argument

    should include the line number of the syntax error
    However these exceptions do not include an offset
    """
    source = """\
foo(bar=baz, bax)
"""
    sourcePath = make_temp_file(source)
    last_line = '            ^\n' if sys.version_info >= (3, 2) else ''
    column = '12:' if sys.version_info >= (3, 2) else ''
    assert_contains_errors(
        sourcePath,
        ["""\
%s:1:%s non-keyword arg after keyword arg
foo(bar=baz, bax)
%s""" % (sourcePath, column, last_line)])


def test_invalidEscape():
    """The invalid escape syntax raises ValueError in Python 2."""
    sourcePath = make_temp_file(r"foo = '\xyz'")
    assert_contains_errors(sourcePath, ("(unicode error) 'unicodeescape' codec can't decode bytes",))


def test_permissionDenied():
    """If the source file is not readable, this is reported on standard error."""
    sourcePath = make_temp_file('')
    os.chmod(sourcePath, 0)
    count, errors = get_errors(sourcePath)
    assert count == 1
    assert errors == [('unexpected_error', sourcePath, "Permission denied")]


def test_frostedWarning():
    """If the source file has a frosted warning, this is reported as a 'flake'."""
    sourcePath = make_temp_file("import foo")
    count, errors = get_errors(sourcePath)
    assert count == 1
    assert errors == [('flake', str(UnusedImport(sourcePath, Node(1), 'foo')))]


@pytest.mark.skipif("PY3")
def test_misencodedFileUTF8():
    """If a source file contains bytes which cannot be decoded, this is reported on stderr."""
    SNOWMAN = chr(0x2603)
    source = ("""\
# coding: ascii
x = "%s"
""" % SNOWMAN).encode('utf-8')
    sourcePath = make_temp_file(source)
    assert_contains_errors(sourcePath, ["%s: problem decoding source\n" % (sourcePath, )])


def test_misencodedFileUTF16():
    """If a source file contains bytes which cannot be decoded, this is reported on stderr."""
    SNOWMAN = chr(0x2603)
    source = ("""\
# coding: ascii
x = "%s"
""" % SNOWMAN).encode('utf-16')
    sourcePath = make_temp_file(source)
    assert_contains_errors(sourcePath, ["%s: problem decoding source\n" % (sourcePath,)])


def test_check_recursive():
    """check_recursive descends into each directory, finding Python files and reporting problems."""
    tempdir = tempfile.mkdtemp()
    os.mkdir(os.path.join(tempdir, 'foo'))
    file1 = os.path.join(tempdir, 'foo', 'bar.py')
    fd = open(file1, 'wb')
    fd.write("import baz\n".encode('ascii'))
    fd.close()
    file2 = os.path.join(tempdir, 'baz.py')
    fd = open(file2, 'wb')
    fd.write("import contraband".encode('ascii'))
    fd.close()
    log = []
    reporter = LoggingReporter(log)
    warnings = check_recursive([tempdir], reporter)
    assert warnings == 2
    assert sorted(log) == sorted([('flake', str(UnusedImport(file1, Node(1), 'baz'))),
                                  ('flake', str(UnusedImport(file2, Node(1), 'contraband')))])



"""frosted/test/test_script.py.

Tests functionality (integration testing) that require starting a full frosted instance against input files

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""

from __future__ import absolute_import, division, print_function, unicode_literals

import os
import shutil
import subprocess
import sys
import tempfile

import frosted
import pytest
from frosted.api import iter_source_code
from frosted.messages import UnusedImport
from pies.overrides import *

from .utils import Node

FROSTED_BINARY = os.path.join(os.path.dirname(frosted.__file__), '..', 'scripts', 'frosted')


def setup_function(function):
    globals()['TEMP_DIR'] = tempfile.mkdtemp()
    globals()['TEMP_FILE_PATH'] = os.path.join(TEMP_DIR, 'temp')


def teardown_function(function):
    shutil.rmtree(TEMP_DIR)


def make_empty_file(*parts):
    assert parts
    fpath = os.path.join(TEMP_DIR, *parts)
    fd = open(fpath, 'a')
    fd.close()
    return fpath


def test_emptyDirectory():
    """There are no Python files in an empty directory."""
    assert list(iter_source_code([TEMP_DIR])) == []


def test_singleFile():
    """If the directory contains one Python file - iter_source_code will find it"""
    childpath = make_empty_file('foo.py')
    assert list(iter_source_code([TEMP_DIR])) == [childpath]


def test_onlyPythonSource():
    """Files that are not Python source files are not included."""
    make_empty_file('foo.pyc')
    assert list(iter_source_code([TEMP_DIR])) == []


def test_recurses():
    """If the Python files are hidden deep down in child directories, we will find them."""
    os.mkdir(os.path.join(TEMP_DIR, 'foo'))
    apath = make_empty_file('foo', 'a.py')
    os.mkdir(os.path.join(TEMP_DIR, 'bar'))
    bpath = make_empty_file('bar', 'b.py')
    cpath = make_empty_file('c.py')
    assert sorted(iter_source_code([TEMP_DIR])) == sorted([apath, bpath, cpath])


def test_multipleDirectories():
    """iter_source_code can be given multiple directories - it will recurse into each of them"""
    foopath = os.path.join(TEMP_DIR, 'foo')
    barpath = os.path.join(TEMP_DIR, 'bar')
    os.mkdir(foopath)
    apath = make_empty_file('foo', 'a.py')
    os.mkdir(barpath)
    bpath = make_empty_file('bar', 'b.py')
    assert sorted(iter_source_code([foopath, barpath])) == sorted([apath, bpath])


def test_explicitFiles():
    """If one of the paths given to iter_source_code is not a directory but a
    file, it will include that in its output."""
    epath = make_empty_file('e.py')
    assert list(iter_source_code([epath])) == [epath]


def run_frosted(paths, stdin=None):
    """Launch a subprocess running frosted."""
    env = native_dict(os.environ)
    env['PYTHONPATH'] = os.pathsep.join(sys.path)
    command = [sys.executable, FROSTED_BINARY]
    command.extend(paths)
    if stdin:
        p = subprocess.Popen(command, env=env, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        (stdout, stderr) = p.communicate(stdin)
    else:
        p = subprocess.Popen(command, env=env, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        (stdout, stderr) = p.communicate()
    rv = p.wait()
    if PY3:
        stdout = stdout.decode('utf-8')
        stderr = stderr.decode('utf-8')

    return (stdout, stderr, rv)


def test_goodFile():
    """When a Python source file is all good, the return code is zero and no

    messages are printed to either stdout or stderr.

    """
    fd = open(TEMP_FILE_PATH, 'a')
    fd.close()
    d = run_frosted([TEMP_FILE_PATH])
    assert d == ('', '', 0)


def test_fileWithFlakes():
    """ When a Python source file has warnings,

    the return code is non-zero and the warnings are printed to stdout

    """
    fd = open(TEMP_FILE_PATH, 'wb')
    fd.write("import contraband\n".encode('ascii'))
    fd.close()
    d = run_frosted([TEMP_FILE_PATH])
    expected = UnusedImport(TEMP_FILE_PATH, Node(1), 'contraband')
    assert d[0].strip() == expected.message.strip()


def test_errors():
    """ When frosted finds errors with the files it's given, (if they don't exist, say),

    then the return code is non-zero and the errors are printed to stderr

    """
    d = run_frosted([TEMP_FILE_PATH, '-r'])
    error_msg = '%s: No such file or directory\n' % (TEMP_FILE_PATH,)
    assert d == ('', error_msg, 1)


def test_readFromStdin():
    """If no arguments are passed to C{frosted} then it reads from stdin."""
    d = run_frosted(['-'], stdin='import contraband'.encode('ascii'))

    expected = UnusedImport('<stdin>', Node(1), 'contraband')
    assert d[0].strip() == expected.message.strip()


@pytest.mark.skipif("version_info >= (3,)")
def test_print_statement_python2():
    d = run_frosted(['-'], stdin='print "Hello, Frosted"'.encode('ascii'))
    assert d == ('', '', 0)



from __future__ import absolute_import, division, print_function, unicode_literals

import textwrap
from collections import namedtuple

from frosted import checker
from pies.overrides import *

PyCF_ONLY_AST = 1024
__all__ = ['flakes', 'Node', 'LoggingReporter']


def flakes(input, *expectedOutputs, **kw):
    tree = compile(textwrap.dedent(input), "<test>", "exec", PyCF_ONLY_AST)
    results = checker.Checker(tree, **kw)
    outputs = [message.type for message in results.messages]
    expectedOutputs = list(expectedOutputs)
    outputs.sort(key=lambda t: t.name)
    expectedOutputs.sort(key=lambda t: t.name)
    assert outputs == expectedOutputs, ('\n'
                                        'for input:\n'
                                        '%s\n'
                                        'expected outputs:\n'
                                        '%r\n'
                                        'but got:\n'
                                        '%s') % (input, expectedOutputs,
                                                 '\n'.join([str(o) for o in results.messages]))
    return results


class Node(namedtuple('Node', ['lineno', 'col_offset'])):
    """A mock AST Node."""

    def __new__(cls, lineno, col_offset=0):
        return super(Node, cls).__new__(cls, lineno, col_offset)


class LoggingReporter(namedtuple('LoggingReporter', ['log'])):
    """A mock Reporter implementation."""

    def flake(self, message):
        self.log.append(('flake', str(message)))

    def unexpected_error(self, filename, message):
        self.log.append(('unexpected_error', filename, message))

    def syntax_error(self, filename, msg, lineno, offset, line):
        self.log.append(('syntax_error', filename, msg, lineno, offset, line))



"""Tests for various Frosted behavior."""

from __future__ import absolute_import, division, print_function, unicode_literals

from sys import version_info

import pytest
from frosted import messages as m
from pies.overrides import *

from .utils import flakes


def test_duplicateArgs():
    flakes('def fu(bar, bar): pass', m.DuplicateArgument)


@pytest.mark.skipif("'todo: this requires finding all assignments in the function body first'")
def test_localReferencedBeforeAssignment():
    flakes('''
    a = 1
    def f():
        a; a=1
    f()
    ''', m.UndefinedName)


def test_redefinedInListComp():
    """Test that shadowing a variable in a list comprehension raises a warning."""
    flakes('''
    a = 1
    [1 for a, b in [(1, 2)]]
    ''', m.RedefinedInListComp)
    flakes('''
    class A:
        a = 1
        [1 for a, b in [(1, 2)]]
    ''', m.RedefinedInListComp)
    flakes('''
    def f():
        a = 1
        [1 for a, b in [(1, 2)]]
    ''', m.RedefinedInListComp)
    flakes('''
    [1 for a, b in [(1, 2)]]
    [1 for a, b in [(1, 2)]]
    ''')
    flakes('''
    for a, b in [(1, 2)]:
        pass
    [1 for a, b in [(1, 2)]]
    ''')


def test_redefinedInGenerator():
    """Test that reusing a variable in a generator does not raise a warning."""
    flakes('''
    a = 1
    (1 for a, b in [(1, 2)])
    ''')
    flakes('''
    class A:
        a = 1
        list(1 for a, b in [(1, 2)])
    ''')
    flakes('''
    def f():
        a = 1
        (1 for a, b in [(1, 2)])
    ''', m.UnusedVariable)
    flakes('''
    (1 for a, b in [(1, 2)])
    (1 for a, b in [(1, 2)])
    ''')
    flakes('''
    for a, b in [(1, 2)]:
        pass
    (1 for a, b in [(1, 2)])
    ''')


@pytest.mark.skipif('''version_info < (2, 7)''')
def test_redefinedInSetComprehension():
    """Test that reusing a variable in a set comprehension does not raise a warning."""
    flakes('''
    a = 1
    {1 for a, b in [(1, 2)]}
    ''')
    flakes('''
    class A:
        a = 1
        {1 for a, b in [(1, 2)]}
    ''')
    flakes('''
    def f():
        a = 1
        {1 for a, b in [(1, 2)]}
    ''', m.UnusedVariable)
    flakes('''
    {1 for a, b in [(1, 2)]}
    {1 for a, b in [(1, 2)]}
    ''')
    flakes('''
    for a, b in [(1, 2)]:
        pass
    {1 for a, b in [(1, 2)]}
    ''')


@pytest.mark.skipif('''version_info < (2, 7)''')
def test_redefinedInDictComprehension():
    """Test that reusing a variable in a dict comprehension does not raise a warning."""
    flakes('''
    a = 1
    {1: 42 for a, b in [(1, 2)]}
    ''')
    flakes('''
    class A:
        a = 1
        {1: 42 for a, b in [(1, 2)]}
    ''')
    flakes('''
    def f():
        a = 1
        {1: 42 for a, b in [(1, 2)]}
    ''', m.UnusedVariable)
    flakes('''
    {1: 42 for a, b in [(1, 2)]}
    {1: 42 for a, b in [(1, 2)]}
    ''')
    flakes('''
    for a, b in [(1, 2)]:
        pass
    {1: 42 for a, b in [(1, 2)]}
    ''')


def test_redefinedFunction():
    """Test that shadowing a function definition with another one raises a warning."""
    flakes('''
    def a(): pass
    def a(): pass
    ''', m.RedefinedWhileUnused)


def test_redefinedClassFunction():
    """Test that shadowing a function definition in a class suite with another one raises a warning."""
    flakes('''
    class A:
        def a(): pass
        def a(): pass
    ''', m.RedefinedWhileUnused)


def test_redefinedIfElseFunction():
    """Test that shadowing a function definition twice in an if and else block does not raise a warning."""
    flakes('''
    if True:
        def a(): pass
    else:
        def a(): pass
    ''')


def test_redefinedIfFunction():
    """Test that shadowing a function definition within an if block raises a warning."""
    flakes('''
    if True:
        def a(): pass
        def a(): pass
    ''', m.RedefinedWhileUnused)


def test_redefinedTryExceptFunction():
    """Test that shadowing a function definition twice in try and except block does not raise a warning."""
    flakes('''
    try:
        def a(): pass
    except Exception:
        def a(): pass
    ''')


def test_redefinedTryFunction():
    """Test that shadowing a function definition within a try block raises a warning."""
    flakes('''
    try:
        def a(): pass
        def a(): pass
    except Exception:
        pass
    ''', m.RedefinedWhileUnused)


def test_redefinedIfElseInListComp():
    """Test that shadowing a variable in a list comprehension in an if and else block does not raise a warning."""
    flakes('''
    if False:
        a = 1
    else:
        [a for a in '12']
    ''')


def test_redefinedElseInListComp():
    """Test that shadowing a variable in a list comprehension in an else (or if) block raises a warning."""
    flakes('''
    if False:
        pass
    else:
        a = 1
        [a for a in '12']
    ''', m.RedefinedInListComp)


def test_functionDecorator():
    """Test that shadowing a function definition with a decorated version of that function does not raise a warning."""
    flakes('''
    from somewhere import somedecorator

    def a(): pass
    a = somedecorator(a)
    ''')


def test_classFunctionDecorator():
    """Test that shadowing a function definition in a class suite with a
    decorated version of that function does not raise a warning.

    """
    flakes('''
    class A:
        def a(): pass
        a = classmethod(a)
    ''')


@pytest.mark.skipif('''version_info < (2, 6)''')
def test_modernProperty():
    flakes("""
    class A:
        @property
        def t():
            pass
        @t.setter
        def t(self, value):
            pass
        @t.deleter
        def t():
            pass
    """)


def test_unaryPlus():
    """Don't die on unary +."""
    flakes('+1')


def test_undefinedBaseClass():
    """If a name in the base list of a class definition is undefined, a warning is emitted."""
    flakes('''
    class foo(foo):
        pass
    ''', m.UndefinedName)


def test_classNameUndefinedInClassBody():
    """If a class name is used in the body of that class's definition and the

    name is not already defined, a warning is emitted.

    """
    flakes('''
    class foo:
        foo
    ''', m.UndefinedName)


def test_classNameDefinedPreviously():
    """If a class name is used in the body of that class's definition and the
    name was previously defined in some other way, no warning is emitted.

    """
    flakes('''
    foo = None
    class foo:
        foo
    ''')


def test_classRedefinition():
    """If a class is defined twice in the same module, a warning is emitted."""
    flakes('''
    class Foo:
        pass
    class Foo:
        pass
    ''', m.RedefinedWhileUnused)


def test_functionRedefinedAsClass():
    """If a function is redefined as a class, a warning is emitted."""
    flakes('''
    def Foo():
        pass
    class Foo:
        pass
    ''', m.RedefinedWhileUnused)


def test_classRedefinedAsFunction():
    """If a class is redefined as a function, a warning is emitted."""
    flakes('''
    class Foo:
        pass
    def Foo():
        pass
    ''', m.RedefinedWhileUnused)


@pytest.mark.skipif("'todo: Too hard to make this warn but other cases stay silent'")
def test_doubleAssignment():
    """If a variable is re-assigned to without being used, no warning is emitted."""
    flakes('''
    x = 10
    x = 20
    ''', m.RedefinedWhileUnused)


def test_doubleAssignmentConditionally():
    """If a variable is re-assigned within a conditional, no warning is emitted."""
    flakes('''
    x = 10
    if True:
        x = 20
    ''')


def test_doubleAssignmentWithUse():
    """If a variable is re-assigned to after being used, no warning is emitted."""
    flakes('''
    x = 10
    y = x * 2
    x = 20
    ''')


def test_comparison():
    """If a defined name is used on either side of any of the six comparison operators, no warning is emitted."""
    flakes('''
    x = 10
    y = 20
    x < y
    x <= y
    x == y
    x != y
    x >= y
    x > y
    ''')


def test_identity():
    """If a defined name is used on either side of an identity test, no warning is emitted."""
    flakes('''
    x = 10
    y = 20
    x is y
    x is not y
    ''')


def test_containment():
    """If a defined name is used on either side of a containment test, no warning is emitted."""
    flakes('''
    x = 10
    y = 20
    x in y
    x not in y
    ''')


def test_loopControl():
    """break and continue statements are supported."""
    flakes('''
    for x in [1, 2]:
        break
    ''')
    flakes('''
    for x in [1, 2]:
        continue
    ''')


def test_ellipsis():
    """Ellipsis in a slice is supported."""
    flakes('''
    [1, 2][...]
    ''')


def test_extendedSlice():
    """Extended slices are supported."""
    flakes('''
    x = 3
    [1, 2][x,:]
    ''')


def test_varAugmentedAssignment():
    """Augmented assignment of a variable is supported.

    We don't care about var refs.

    """
    flakes('''
    foo = 0
    foo += 1
    ''')


def test_attrAugmentedAssignment():
    """Augmented assignment of attributes is supported.

    We don't care about attr refs.

    """
    flakes('''
    foo = None
    foo.bar += foo.baz
    ''')


def test_unusedVariable():
    """Warn when a variable in a function is assigned a value that's never used."""
    flakes('''
    def a():
        b = 1
    ''', m.UnusedVariable)


def test_unusedVariableAsLocals():
    """Using locals() it is perfectly valid to have unused variables."""
    flakes('''
    def a():
        b = 1
        return locals()
    ''')


def test_unusedVariableNoLocals():
    """Using locals() in wrong scope should not matter."""
    flakes('''
    def a():
        locals()
        def a():
            b = 1
            return
    ''', m.UnusedVariable)


def test_assignToGlobal():
    """Assigning to a global and then not using that global is perfectly
    acceptable.

    Do not mistake it for an unused local variable.

    """
    flakes('''
    b = 0
    def a():
        global b
        b = 1
    ''')


@pytest.mark.skipif('''version_info < (3,)''')
def test_assignToNonlocal():
    """Assigning to a nonlocal and then not using that binding is perfectly
    acceptable.

    Do not mistake it for an unused local variable.

    """
    flakes('''
    b = b'0'
    def a():
        nonlocal b
        b = b'1'
    ''')


def test_assignToMember():
    """Assigning to a member of another object and then not using that member
    variable is perfectly acceptable.

    Do not mistake it for an unused local variable.

    """
    # XXX: Adding this test didn't generate a failure. Maybe not
    # necessary?
    flakes('''
    class b:
        pass
    def a():
        b.foo = 1
    ''')


def test_assignInForLoop():
    """Don't warn when a variable in a for loop is assigned to but not used."""
    flakes('''
    def f():
        for i in range(10):
            pass
    ''')


def test_assignInListComprehension():
    """Don't warn when a variable in a list comprehension is assigned to but not used."""
    flakes('''
    def f():
        [None for i in range(10)]
    ''')


def test_generatorExpression():
    """Don't warn when a variable in a generator expression is assigned to but not used."""
    flakes('''
    def f():
        (None for i in range(10))
    ''')


def test_assignmentInsideLoop():
    """Don't warn when a variable assignment occurs lexically after its use."""
    flakes('''
    def f():
        x = None
        for i in range(10):
            if i > 2:
                return x
            x = i * 2
    ''')


def test_tupleUnpacking():
    """Don't warn when a variable included in tuple unpacking is unused.

    It's very common for variables in a tuple unpacking assignment to be unused in good Python code, so warning will
    only create false positives.

    """
    flakes('''
    def f():
        (x, y) = 1, 2
    ''')


def test_listUnpacking():
    """Don't warn when a variable included in list unpacking is unused."""
    flakes('''
    def f():
        [x, y] = [1, 2]
    ''')


def test_closedOver():
    """Don't warn when the assignment is used in an inner function."""
    flakes('''
    def barMaker():
        foo = 5
        def bar():
            return foo
        return bar
    ''')


def test_doubleClosedOver():
    """Don't warn when the assignment is used in an inner function, even if
    that inner function itself is in an inner function."""
    flakes('''
    def barMaker():
        foo = 5
        def bar():
            def baz():
                return foo
        return bar
    ''')


def test_tracebackhideSpecialVariable():
    """Do not warn about unused local variable __tracebackhide__, which is a
    special variable for py.test."""
    flakes("""
        def helper():
            __tracebackhide__ = True
    """)


def test_ifexp():
    """Test C{foo if bar else baz} statements."""
    flakes("a = 'moo' if True else 'oink'")
    flakes("a = foo if True else 'oink'", m.UndefinedName)
    flakes("a = 'moo' if True else bar", m.UndefinedName)


def test_withStatementNoNames():
    """No warnings are emitted for using inside or after a nameless  statement a name defined beforehand."""
    flakes('''
    from __future__ import with_statement
    bar = None
    with open("foo"):
        bar
    bar
    ''')


def test_withStatementSingleName():
    """No warnings are emitted for using a name defined by a  statement within the suite or afterwards."""
    flakes('''
    from __future__ import with_statement
    with open('foo') as bar:
        bar
    bar
    ''')


def test_withStatementAttributeName():
    """No warnings are emitted for using an attribute as the target of a  statement."""
    flakes('''
    from __future__ import with_statement
    import foo
    with open('foo') as foo.bar:
        pass
    ''')


def test_withStatementSubscript():
    """No warnings are emitted for using a subscript as the target of a  statement."""
    flakes('''
    from __future__ import with_statement
    import foo
    with open('foo') as foo[0]:
        pass
    ''')


def test_withStatementSubscriptUndefined():
    """An undefined name warning is emitted if the subscript used as the target of a with statement is not defined."""
    flakes('''
    from __future__ import with_statement
    import foo
    with open('foo') as foo[bar]:
        pass
    ''', m.UndefinedName)


def test_withStatementTupleNames():
    """No warnings are emitted for using any of the tuple of names defined

    by a statement within the suite or afterwards.

    """
    flakes('''
    from __future__ import with_statement
    with open('foo') as (bar, baz):
        bar, baz
    bar, baz
    ''')


def test_withStatementListNames():
    """No warnings are emitted for using any of the list of names defined by

    a statement within the suite or afterwards.

    """
    flakes('''
    from __future__ import with_statement
    with open('foo') as [bar, baz]:
        bar, baz
    bar, baz
    ''')


def test_withStatementComplicatedTarget():
    """ If the target of a  statement uses any or all of the valid forms
    for that part of the grammar
    (See: http://docs.python.org/reference/compound_stmts.html#the-with-statement),
    the names involved are checked both for definedness and any bindings
    created are respected in the suite of the statement and afterwards.

    """
    flakes('''
    from __future__ import with_statement
    c = d = e = g = h = i = None
    with open('foo') as [(a, b), c[d], e.f, g[h:i]]:
        a, b, c, d, e, g, h, i
    a, b, c, d, e, g, h, i
    ''')


def test_withStatementSingleNameUndefined():
    """An undefined name warning is emitted if the name first defined by a statement is used before the  statement."""
    flakes('''
    from __future__ import with_statement
    bar
    with open('foo') as bar:
        pass
    ''', m.UndefinedName)


def test_withStatementTupleNamesUndefined():
    """ An undefined name warning is emitted if a name first defined by a the
    tuple-unpacking form of the  statement is used before the statement.

    """
    flakes('''
    from __future__ import with_statement
    baz
    with open('foo') as (bar, baz):
        pass
    ''', m.UndefinedName)


def test_withStatementSingleNameRedefined():
    """A redefined name warning is emitted if a name bound by an import is
    rebound by the name defined by a  statement.

    """
    flakes('''
    from __future__ import with_statement
    import bar
    with open('foo') as bar:
        pass
    ''', m.RedefinedWhileUnused)


def test_withStatementTupleNamesRedefined():
    """ A redefined name warning is emitted if a name bound by an import is
    rebound by one of the names defined by the tuple-unpacking form of a
    statement.

    """
    flakes('''
    from __future__ import with_statement
    import bar
    with open('foo') as (bar, baz):
        pass
    ''', m.RedefinedWhileUnused)


def test_withStatementUndefinedInside():
    """An undefined name warning is emitted if a name is used inside the body
    of a  statement without first being bound.

    """
    flakes('''
    from __future__ import with_statement
    with open('foo') as bar:
        baz
    ''', m.UndefinedName)


def test_withStatementNameDefinedInBody():
    """A name defined in the body of a  statement can be used after the body ends without warning."""
    flakes('''
    from __future__ import with_statement
    with open('foo') as bar:
        baz = 10
    baz
    ''')


def test_withStatementUndefinedInExpression():
    """An undefined name warning is emitted if a name in the I{test} expression of a  statement is undefined."""
    flakes('''
    from __future__ import with_statement
    with bar as baz:
        pass
    ''', m.UndefinedName)

    flakes('''
    from __future__ import with_statement
    with bar as bar:
        pass
    ''', m.UndefinedName)


@pytest.mark.skipif('''version_info < (2, 7)''')
def test_dictComprehension():
    """Dict comprehensions are properly handled."""
    flakes('''
    a = {1: x for x in range(10)}
    ''')


@pytest.mark.skipif('''version_info < (2, 7)''')
def test_setComprehensionAndLiteral():
    """Set comprehensions are properly handled."""
    flakes('''
    a = {1, 2, 3}
    b = {x for x in range(10)}
    ''')


def test_exceptionUsedInExcept():
    as_exc = ', ' if version_info < (2, 6) else ' as '
    flakes('''
    try: pass
    except Exception%se: e
    ''' % as_exc)

    flakes('''
    def download_review():
        try: pass
        except Exception%se: e
    ''' % as_exc)


def test_exceptWithoutNameInFunction():
    """Don't issue false warning when an unnamed exception is used.

    Previously, there would be a false warning, but only when the try..except was in a function

    """
    flakes('''
    import tokenize
    def foo():
        try: pass
        except tokenize.TokenError: pass
    ''')


def test_exceptWithoutNameInFunctionTuple():
    """Don't issue false warning when an unnamed exception is used.

    This example catches a tuple of exception types.

    """
    flakes('''
    import tokenize
    def foo():
        try: pass
        except (tokenize.TokenError, IndentationError): pass
    ''')


def test_augmentedAssignmentImportedFunctionCall():
    """Consider a function that is called on the right part of an augassign operation to be used."""
    flakes('''
    from foo import bar
    baz = 0
    baz += bar()
    ''')


@pytest.mark.skipif('''version_info < (3, 3)''')
def test_yieldFromUndefined():
    """Test C{yield from} statement."""
    flakes('''
    def bar():
        yield from foo()
    ''', m.UndefinedName)


def test_bareExcept():
    """
        Issue a warning when using bare except:.
    """
    flakes('''
        try:
            pass
        except:
            pass
        ''', m.BareExcept)



from __future__ import absolute_import, division, print_function, unicode_literals

from sys import version_info

import pytest
from frosted import messages as m
from pies.overrides import *

from .utils import flakes


@pytest.mark.skipif("version_info >= (3,)")
def test_return():
    flakes('''
    class a:
        def b():
            for x in a.c:
                if x:
                    yield x
            return a
    ''', m.ReturnWithArgsInsideGenerator)


@pytest.mark.skipif("version_info >= (3,)")
def test_returnNone():
    flakes('''
    def a():
        yield 12
        return None
    ''', m.ReturnWithArgsInsideGenerator)


@pytest.mark.skipif("version_info >= (3,)")
def test_returnYieldExpression():
    flakes('''
    def a():
        b = yield a
        return b
    ''', m.ReturnWithArgsInsideGenerator)


@pytest.mark.skipif("version_info >= (3,)")
def test_return_with_args_inside_generator_not_duplicated():
    # doubly nested - should still only complain once
    flakes('''
    def f0():
        def f1():
            yield None
            return None
    ''', m.ReturnWithArgsInsideGenerator)

    # triple nested - should still only complain once
    flakes('''
    def f0():
        def f1():
            def f2():
                yield None
                return None
    ''', m.ReturnWithArgsInsideGenerator)


@pytest.mark.skipif("version_info >= (3,)")
def test_no_false_positives_for_return_inside_generator():
    # doubly nested - should still only complain once
    flakes('''
    def f():
        def g():
            yield None
        return g
    ''')



"""frosted/api.py.

Defines the api for the command-line frosted utility

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR

"""
from __future__ import unicode_literals

import os
import sys

from frosted import reporter as modReporter
from frosted import checker, settings
from pies.overrides import *

import _ast

__all__ = ['check', 'check_path', 'check_recursive', 'iter_source_code', 'main']


def check(codeString, filename, reporter=modReporter.Default, **setting_overrides):
    """Check the Python source given by codeString for unfrosted flakes."""
    active_settings = settings.default.copy()
    active_settings.update(setting_overrides)

    if filename in active_settings.get('skip', []) or filename.split('/')[-1] in active_settings.get('skip', []):
        sys.stderr.write("WARNING: {0} was skipped as it's listed in 'skip' setting\n".format(filename))
        return 0

    # First, compile into an AST and handle syntax errors.
    try:
        tree = compile(codeString, filename, "exec", _ast.PyCF_ONLY_AST)
    except SyntaxError:
        value = sys.exc_info()[1]
        msg = value.args[0]

        (lineno, offset, text) = value.lineno, value.offset, value.text

        # If there's an encoding problem with the file, the text is None.
        if text is None:
            # Avoid using msg, since for the only known case, it contains a
            # bogus message that claims the encoding the file declared was
            # unknown.
            reporter.unexpected_error(filename, 'problem decoding source')
        else:
            reporter.syntax_error(filename, msg, lineno, offset, text)
        return 1
    except Exception:
        reporter.unexpected_error(filename, 'problem decoding source')
        return 1
    # Okay, it's syntactically valid.  Now check it.
    w = checker.Checker(tree, filename, None, **active_settings)
    w.messages.sort(key=lambda m: m.lineno)
    for warning in w.messages:
        reporter.flake(warning)
    return len(w.messages)


def check_path(filename, reporter=modReporter.Default, **setting_overrides):
    """Check the given path, printing out any warnings detected."""
    try:
        with open(filename, 'U') as f:
            codestr = f.read() + '\n'
    except UnicodeError:
        reporter.unexpected_error(filename, 'problem decoding source')
        return 1
    except IOError:
        msg = sys.exc_info()[1]
        reporter.unexpected_error(filename, msg.args[1])
        return 1
    return check(codestr, filename, reporter, **setting_overrides)


def iter_source_code(paths):
    """Iterate over all Python source files defined in paths."""
    for path in paths:
        if os.path.isdir(path):
            for dirpath, dirnames, filenames in os.walk(path):
                for filename in filenames:
                    if filename.endswith('.py'):
                        yield os.path.join(dirpath, filename)
        else:
            yield path


def check_recursive(paths, reporter=modReporter.Default, **setting_overrides):
    """Recursively check all source files defined in paths."""
    warnings = 0
    for source_path in iter_source_code(paths):
        warnings += check_path(source_path, reporter, **setting_overrides)
    return warnings



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

import sys

from frosted import messages as m
from pies.overrides import *

from .utils import flakes


def test_ok():
    flakes('''
    def foo(a):
        pass
    foo(5)
    ''')

    flakes('''
    def foo(a, b=2):
        pass
    foo(5, b=1)
    ''')


def test_noCheckDecorators():
    flakes('''
    def decorator(f):
        return f
    @decorator
    def foo():
        pass
    foo(42)
    ''')


def test_tooManyArguments():
    flakes('''
    def foo():
        pass
    foo(5)
    ''', m.TooManyArguments)
    flakes('''
    def foo(a, b):
        pass
    foo(5, 6, 7)
    ''', m.TooManyArguments)


def test_tooManyArgumentsVarargs():
    flakes('''
    def foo(a, *args):
        pass
    foo(1, 2, 3)
    ''')


def test_unexpectedArgument():
    flakes('''
    def foo(a):
        pass
    foo(1, b=3)
    ''', m.UnexpectedArgument)

    flakes('''
    def foo(a, *args):
        pass
    foo(1, b=3)
    ''', m.UnexpectedArgument)

    flakes('''
    def foo(a, **kwargs):
        pass
    foo(1, b=3)
    ''')


def test_multipleValuesForArgument():
    flakes('''
    def foo(a):
        pass
    foo(5, a=5)
    ''', m.MultipleValuesForArgument)


def test_tooFewArguments():
    flakes('''
    def foo(a):
        pass
    foo()
    ''', m.TooFewArguments)

    flakes('''
    def foo(a):
        pass
    foo(*[])
    ''')

    flakes('''
    def foo(a):
        pass
    foo(**{})
    ''')


def test_tooFewArgumentsVarArgs():
    flakes('''
    def foo(a, b, *args):
        pass
    foo(1)
    ''', m.TooFewArguments)


if PY3:
    def test_kwOnlyArguments():
        flakes('''
        def foo(a, *, b=0):
            pass
        foo(5, b=2)
        ''')

        flakes('''
        def foo(a, *, b=0):
            pass
        foo(5)
        ''')

        flakes('''
        def foo(a, *, b):
            pass
        foo(5, b=2)
        ''')

        flakes('''
        def foo(a, *, b):
            pass
        foo(5, **{})
        ''')

        flakes('''
        def foo(a, *, b):
            pass
        foo(1)
        ''', m.NeedKwOnlyArgument)

        flakes('''
        def foo(a, *args, b):
            pass
        foo(1, 2, 3, 4)
        ''', m.NeedKwOnlyArgument)
elif PY2:
    def test_compoundArguments():
        flakes('''
        def foo(a, (b, c)):
            pass
        foo(1, [])''')

        flakes('''
        def foo(a, (b, c)):
            pass
        foo(1, 2, 3)''', m.TooManyArguments)

        flakes('''
        def foo(a, (b, c)):
            pass
        foo(1)''', m.TooFewArguments)

        flakes('''
        def foo(a, (b, c)):
            pass
        foo(1, b=2, c=3)''', m.UnexpectedArgument)




from __future__ import absolute_import, division, print_function, unicode_literals

from sys import version_info

import pytest
from frosted import messages as m
from pies.overrides import *

from .utils import flakes


def test_unusedImport():
    flakes('import fu, bar', m.UnusedImport, m.UnusedImport)
    flakes('from baz import fu, bar', m.UnusedImport, m.UnusedImport)


def test_aliasedImport():
    flakes('import fu as FU, bar as FU',
                m.RedefinedWhileUnused, m.UnusedImport)
    flakes('from moo import fu as FU, bar as FU',
                m.RedefinedWhileUnused, m.UnusedImport)


def test_usedImport():
    flakes('import fu; print(fu)')
    flakes('from baz import fu; print(fu)')
    flakes('import fu; del fu')


def test_redefinedWhileUnused():
    flakes('import fu; fu = 3', m.RedefinedWhileUnused)
    flakes('import fu; fu, bar = 3', m.RedefinedWhileUnused)
    flakes('import fu; [fu, bar] = 3', m.RedefinedWhileUnused)


def test_redefinedIf():
    """Test that importing a module twice within an if block does raise a warning."""
    flakes('''
    i = 2
    if i==1:
        import os
        import os
    os.path''', m.RedefinedWhileUnused)


def test_redefinedIfElse():
    """Test that importing a module twice in if and else blocks does not raise a warning."""
    flakes('''
    i = 2
    if i==1:
        import os
    else:
        import os
    os.path''')


def test_redefinedTry():
    """Test that importing a module twice in an try block does raise a warning."""
    flakes('''
    try:
        import os
        import os
    except Exception:
        pass
    os.path''', m.RedefinedWhileUnused)


def test_redefinedTryExcept():
    """Test that importing a module twice in an try and except block does not raise a warning."""
    flakes('''
    try:
        import os
    except Exception:
        import os
    os.path''')


def test_redefinedTryNested():
    """Test that importing a module twice using a nested try/except and if blocks does not issue a warning."""
    flakes('''
    try:
        if True:
            if True:
                import os
    except Exception:
        import os
    os.path''')


def test_redefinedTryExceptMulti():
    flakes("""
    try:
        from aa import mixer
    except AttributeError:
        from bb import mixer
    except RuntimeError:
        from cc import mixer
    except Exception:
        from dd import mixer
    mixer(123)
    """)


def test_redefinedTryElse():
    flakes("""
    try:
        from aa import mixer
    except ImportError:
        pass
    else:
        from bb import mixer
    mixer(123)
    """, m.RedefinedWhileUnused)


def test_redefinedTryExceptElse():
    flakes("""
    try:
        import funca
    except ImportError:
        from bb import funca
        from bb import funcb
    else:
        from bbb import funcb
    print(funca, funcb)
    """)


def test_redefinedTryExceptFinally():
    flakes("""
    try:
        from aa import a
    except ImportError:
        from bb import a
    finally:
        a = 42
    print(a)
    """)


def test_redefinedTryExceptElseFinally():
    flakes("""
    try:
        import b
    except ImportError:
        b = Ellipsis
        from bb import a
    else:
        from aa import a
    finally:
        a = 42
    print(a, b)
    """)


def test_redefinedByFunction():
    flakes('''
    import fu
    def fu():
        pass
    ''', m.RedefinedWhileUnused)


def test_redefinedInNestedFunction():
    """Test that shadowing a global name with a nested function definition generates a warning."""
    flakes('''
    import fu
    def bar():
        def baz():
            def fu():
                pass
    ''', m.RedefinedWhileUnused, m.UnusedImport)


def test_redefinedByClass():
    flakes('''
    import fu
    class fu:
        pass
    ''', m.RedefinedWhileUnused)


def test_redefinedBySubclass():
    """If an imported name is redefined by a class statement

    which also uses that name in the bases list, no warning is emitted.

    """
    flakes('''
    from fu import bar
    class bar(bar):
        pass
    ''')


def test_redefinedInClass():
    """Test that shadowing a global with a class attribute does not produce a warning."""
    flakes('''
    import fu
    class bar:
        fu = 1
    print(fu)
    ''')


def test_usedInFunction():
    flakes('''
    import fu
    def fun():
        print(fu)
    ''')


def test_shadowedByParameter():
    flakes('''
    import fu
    def fun(fu):
        print(fu)
    ''', m.UnusedImport)

    flakes('''
    import fu
    def fun(fu):
        print(fu)
    print(fu)
    ''')


def test_newAssignment():
    flakes('fu = None')


def test_usedInGetattr():
    flakes('import fu; fu.bar.baz')
    flakes('import fu; "bar".fu.baz', m.UnusedImport)


def test_usedInSlice():
    flakes('import fu; print(fu.bar[1:])')


def test_usedInIfBody():
    flakes('''
    import fu
    if True: print(fu)
    ''')


def test_usedInIfConditional():
    flakes('''
    import fu
    if fu: pass
    ''')


def test_usedInElifConditional():
    flakes('''
    import fu
    if False: pass
    elif fu: pass
    ''')


def test_usedInElse():
    flakes('''
    import fu
    if False: pass
    else: print(fu)
    ''')


def test_usedInCall():
    flakes('import fu; fu.bar()')


def test_usedInClass():
    flakes('''
    import fu
    class bar:
        bar = fu
    ''')


def test_usedInClassBase():
    flakes('''
    import fu
    class bar(object, fu.baz):
        pass
    ''')


def test_notUsedInNestedScope():
    flakes('''
    import fu
    def bleh():
        pass
    print(fu)
    ''')


def test_usedInFor():
    flakes('''
    import fu
    for bar in range(9):
        print(fu)
    ''')


def test_usedInForElse():
    flakes('''
    import fu
    for bar in range(10):
        pass
    else:
        print(fu)
    ''')


def test_redefinedByFor():
    flakes('''
    import fu
    for fu in range(2):
        pass
    ''', m.RedefinedWhileUnused)


def test_shadowedByFor():
    """Test that shadowing a global name with a for loop variable generates a warning."""
    flakes('''
    import fu
    fu.bar()
    for fu in ():
        pass
    ''', m.ImportShadowedByLoopVar)


def test_shadowedByForDeep():
    """Test that shadowing a global name with a for loop variable nested in a tuple unpack generates a warning."""
    flakes('''
    import fu
    fu.bar()
    for (x, y, z, (a, b, c, (fu,))) in ():
        pass
    ''', m.ImportShadowedByLoopVar)


def test_usedInReturn():
    flakes('''
    import fu
    def fun():
        return fu
    ''')


def test_usedInOperators():
    flakes('import fu; 3 + fu.bar')
    flakes('import fu; 3 % fu.bar')
    flakes('import fu; 3 - fu.bar')
    flakes('import fu; 3 * fu.bar')
    flakes('import fu; 3 ** fu.bar')
    flakes('import fu; 3 / fu.bar')
    flakes('import fu; 3 // fu.bar')
    flakes('import fu; -fu.bar')
    flakes('import fu; ~fu.bar')
    flakes('import fu; 1 == fu.bar')
    flakes('import fu; 1 | fu.bar')
    flakes('import fu; 1 & fu.bar')
    flakes('import fu; 1 ^ fu.bar')
    flakes('import fu; 1 >> fu.bar')
    flakes('import fu; 1 << fu.bar')


def test_usedInAssert():
    flakes('import fu; assert fu.bar')


def test_usedInSubscript():
    flakes('import fu; fu.bar[1]')


def test_usedInLogic():
    flakes('import fu; fu and False')
    flakes('import fu; fu or False')
    flakes('import fu; not fu.bar')


def test_usedInList():
    flakes('import fu; [fu]')


def test_usedInTuple():
    flakes('import fu; (fu,)')


def test_usedInTry():
    flakes('''
    import fu
    try: fu
    except Exception: pass
    ''')


def test_usedInExcept():
    flakes('''
    import fu
    try: fu
    except Exception: pass
    ''')


def test_redefinedByExcept():
    as_exc = ', ' if version_info < (2, 6) else ' as '
    flakes('''
    import fu
    try: pass
    except Exception%sfu: pass
    ''' % as_exc, m.RedefinedWhileUnused)


def test_usedInRaise():
    flakes('''
    import fu
    raise fu.bar
    ''')


def test_usedInYield():
    flakes('''
    import fu
    def gen():
        yield fu
    ''')


def test_usedInDict():
    flakes('import fu; {fu:None}')
    flakes('import fu; {1:fu}')


def test_usedInParameterDefault():
    flakes('''
    import fu
    def f(bar=fu):
        pass
    ''')


def test_usedInAttributeAssign():
    flakes('import fu; fu.bar = 1')


def test_usedInKeywordArg():
    flakes('import fu; fu.bar(stuff=fu)')


def test_usedInAssignment():
    flakes('import fu; bar=fu')
    flakes('import fu; n=0; n+=fu')


def test_usedInListComp():
    flakes('import fu; [fu for _ in range(1)]')
    flakes('import fu; [1 for _ in range(1) if fu]')


def test_redefinedByListComp():
    flakes('import fu; [1 for fu in range(1)]', m.RedefinedWhileUnused)


def test_usedInTryFinally():
    flakes('''
    import fu
    try: pass
    finally: fu
    ''')

    flakes('''
    import fu
    try: fu
    finally: pass
    ''')


def test_usedInWhile():
    flakes('''
    import fu
    while 0:
        fu
    ''')

    flakes('''
    import fu
    while fu: pass
    ''')


def test_usedInGlobal():
    flakes('''
    import fu
    def f(): global fu
    ''', m.UnusedImport)


@pytest.mark.skipif("version_info >= (3,)")
def test_usedInBackquote():
    flakes('import fu; `fu`')


def test_usedInExec():
    if version_info < (3,):
        exec_stmt = 'exec "print 1" in fu.bar'
    else:
        exec_stmt = 'exec("print(1)", fu.bar)'
    flakes('import fu; %s' % exec_stmt)


def test_usedInLambda():
    flakes('import fu; lambda: fu')


def test_shadowedByLambda():
    flakes('import fu; lambda fu: fu', m.UnusedImport)


def test_usedInSliceObj():
    flakes('import fu; "meow"[::fu]')


def test_unusedInNestedScope():
    flakes('''
    def bar():
        import fu
    fu
    ''', m.UnusedImport, m.UndefinedName)


def test_methodsDontUseClassScope():
    flakes('''
    class bar:
        import fu
        def fun():
            fu
    ''', m.UnusedImport, m.UndefinedName)


def test_nestedFunctionsNestScope():
    flakes('''
    def a():
        def b():
            fu
        import fu
    ''')


def test_nestedClassAndFunctionScope():
    flakes('''
    def a():
        import fu
        class b:
            def c():
                print(fu)
    ''')


def test_importStar():
    flakes('from fu import *', m.ImportStarUsed)


def test_packageImport():
    """If a dotted name is imported and used, no warning is reported."""
    flakes('''
    import fu.bar
    fu.bar
    ''')


def test_unusedPackageImport():
    """If a dotted name is imported and not used, an unused import warning is reported."""
    flakes('import fu.bar', m.UnusedImport)


def test_duplicateSubmoduleImport():
    """If a submodule of a package is imported twice, an unused

    import warning and a redefined while unused warning are reported.

    """
    flakes('''
    import fu.bar, fu.bar
    fu.bar
    ''', m.RedefinedWhileUnused)
    flakes('''
    import fu.bar
    import fu.bar
    fu.bar
    ''', m.RedefinedWhileUnused)


def test_differentSubmoduleImport():
    """If two different submodules of a package are imported,

    no duplicate import warning is reported for the package.

    """
    flakes('''
    import fu.bar, fu.baz
    fu.bar, fu.baz
    ''')
    flakes('''
    import fu.bar
    import fu.baz
    fu.bar, fu.baz
    ''')


def test_assignRHSFirst():
    flakes('import fu; fu = fu')
    flakes('import fu; fu, bar = fu')
    flakes('import fu; [fu, bar] = fu')
    flakes('import fu; fu += fu')


def test_tryingMultipleImports():
    flakes('''
    try:
        import fu
    except ImportError:
        import bar as fu
    fu
    ''')


def test_nonGlobalDoesNotRedefine():
    flakes('''
    import fu
    def a():
        fu = 3
        return fu
    fu
    ''')


def test_functionsRunLater():
    flakes('''
    def a():
        fu
    import fu
    ''')


def test_functionNamesAreBoundNow():
    flakes('''
    import fu
    def fu():
        fu
    fu
    ''', m.RedefinedWhileUnused)


def test_ignoreNonImportRedefinitions():
    flakes('a = 1; a = 2')


@pytest.mark.skipif("'todo'")
def test_importingForImportError():
    flakes('''
    try:
        import fu
    except ImportError:
        pass
    ''')


@pytest.mark.skipif("'todo: requires evaluating attribute access'")
def test_importedInClass():
    """Imports in class scope can be used through."""
    flakes('''
    class c:
        import i
        def __init__():
            i
    ''')


def test_futureImport():
    """__future__ is special."""
    flakes('from __future__ import division')
    flakes('''
    "docstring is allowed before future import"
    from __future__ import division
    ''')


def test_futureImportFirst():
    """__future__ imports must come before anything else."""
    flakes('''
    x = 5
    from __future__ import division
    ''', m.LateFutureImport)
    flakes('''
    from foo import bar
    from __future__ import division
    bar
    ''', m.LateFutureImport)


def test_ignoredInFunction():
    """An C{__all__} definition does not suppress unused import warnings in a function scope."""
    flakes('''
    def foo():
        import bar
        __all__ = ["bar"]
    ''', m.UnusedImport, m.UnusedVariable)


def test_ignoredInClass():
    """An C{__all__} definition does not suppress unused import warnings in a class scope."""
    flakes('''
    class foo:
        import bar
        __all__ = ["bar"]
    ''', m.UnusedImport)


def test_warningSuppressed():
    """If a name is imported and unused but is named in C{__all__}, no warning is reported."""
    flakes('''
    import foo
    __all__ = ["foo"]
    ''')


def test_unrecognizable():
    """If C{__all__} is defined in a way that can't be recognized statically, it is ignored."""
    flakes('''
    import foo
    __all__ = ["f" + "oo"]
    ''', m.UnusedImport)
    flakes('''
    import foo
    __all__ = [] + ["foo"]
    ''', m.UnusedImport)


def test_unboundExported():
    """If C{__all__} includes a name which is not bound, a warning is emitted."""
    flakes('''
    __all__ = ["foo"]
    ''', m.UndefinedExport)

    # Skip this in __init__.py though, since the rules there are a little
    # different.
    for filename in ["foo/__init__.py", "__init__.py"]:
        flakes('''
        __all__ = ["foo"]
        ''', filename=filename)


def test_importStarExported():
    """Do not report undefined if import * is used"""
    flakes('''
    from foolib import *
    __all__ = ["foo"]
    ''', m.ImportStarUsed)


def test_usedInGenExp():
    """Using a global in a generator expression results in no warnings."""
    flakes('import fu; (fu for _ in range(1))')
    flakes('import fu; (1 for _ in range(1) if fu)')


def test_redefinedByGenExp():
    """ Re-using a global name as the loop variable for a generator

    expression results in a redefinition warning.

    """
    flakes('import fu; (1 for fu in range(1))', m.RedefinedWhileUnused, m.UnusedImport)


def test_usedAsDecorator():
    """Using a global name in a decorator statement results in no warnings, but

    using an undefined name in a decorator statement results in an undefined
    name warning.

    """
    flakes('''
    from interior import decorate
    @decorate
    def f():
        return "hello"
    ''')

    flakes('''
    from interior import decorate
    @decorate('value')
    def f():
        return "hello"
    ''')

    flakes('''
    @decorate
    def f():
        return "hello"
    ''', m.UndefinedName)


def test_usedAsClassDecorator():
    """Using an imported name as a class decorator results in no warnings

    but using an undefined name as a class decorator results in an undefined name warning.

    """
    flakes('''
    from interior import decorate
    @decorate
    class foo:
        pass
    ''')

    flakes('''
    from interior import decorate
    @decorate("foo")
    class bar:
        pass
    ''')

    flakes('''
    @decorate
    class foo:
        pass
    ''', m.UndefinedName)



"""frosted/reporter.py.

Defines how errors found by frosted should be displayed to the user

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR

"""
from __future__ import absolute_import, division, print_function, unicode_literals

import re
import sys
from collections import namedtuple

from pies.overrides import *


class Reporter(namedtuple('Reporter', ('stdout', 'stderr'))):
    """Formats the results of frosted checks and then presents them to the user."""

    def unexpected_error(self, filename, msg):
        """Output an unexpected_error specific to the provided filename."""
        self.stderr.write("%s: %s\n" % (filename, msg))

    def syntax_error(self, filename, msg, lineno, offset, text):
        """Output a syntax_error specific to the provided filename."""
        line = text.splitlines()[-1]
        if offset is not None:
            offset = offset - (len(text) - len(line))
            self.stderr.write('%s:%d:%d: %s\n' % (filename, lineno, offset, msg))
        else:
            self.stderr.write('%s:%d: %s\n' % (filename, lineno, msg))
        self.stderr.write(str(line))
        self.stderr.write('\n')
        if offset is not None:
            self.stderr.write(re.sub(r'\S', ' ', line[:offset]) + "^\n")

    def flake(self, message):
        """Print an error message to stdout."""
        self.stdout.write(str(message))
        self.stdout.write('\n')

Default = Reporter(sys.stdout, sys.stderr)



"""frosted/settings.py.

Defines how the default settings for frosted should be loaded

(First from the default setting dictionary at the top of the file, then overridden by any settings
 in ~/.frosted.conf if there are any)

Copyright (C) 2013  Timothy Edmund Crosley

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""
from __future__ import absolute_import, division, print_function, unicode_literals

import os

from pies.overrides import *

MAX_CONFIG_SEARCH_DEPTH = 25 # The number of parent directories frosted will look for a config file within

# Note that none of these lists must be complete as they are simply fallbacks for when included auto-detection fails.
default = {'skip': ['__init__.py', ],
           'ignore_frosted_errors': [],
           'verbose': False}

try:
    import configparser
except ImportError:
    import ConfigParser as configparser

editor_config_file = os.path.expanduser('~/.editorconfig')
tries = 0
current_directory = os.getcwd()
while current_directory and tries < MAX_CONFIG_SEARCH_DEPTH:
    potential_path = os.path.join(current_directory, ".editorconfig")
    if os.path.exists(potential_path):
        editor_config_file = potential_path
        break

    current_directory = os.path.split(current_directory)[0]
    tries += 1

if os.path.exists(editor_config_file):
    with open(editor_config_file) as config_file:
        line = "\n"
        last_position = config_file.tell()
        while line:
            line = config_file.readline()
            if "[" in line:
                config_file.seek(last_position)
                break
            last_position = config_file.tell()

        config = configparser.SafeConfigParser()
        config.readfp(config_file)
        settings = {}
        if config.has_section('*'):
            settings.update(dict(config.items('*')))
        if config.has_section('*.py'):
            settings.update(dict(config.items('*.py')))
        if config.has_section('**.py'):
            settings.update(dict(config.items('**.py')))

        for key, value in settings.items():
            existing_value_type = type(default.get(key, ''))
            if existing_value_type in (list, tuple):
                default[key.lower()] = value.split(",")
            else:
                default[key.lower()] = existing_value_type(value)

frosted_config_file = os.path.expanduser('~/.frosted.cfg')
tries = 0
current_directory = os.getcwd()
while current_directory and tries < MAX_CONFIG_SEARCH_DEPTH:
    potential_path = os.path.join(current_directory, ".frosted.cfg")
    if os.path.exists(potential_path):
        frosted_config_file = potential_path
        break

    current_directory = os.path.split(current_directory)[0]
    tries += 1

if os.path.exists(frosted_config_file):
    with open(frosted_config_file) as config_file:
        config = configparser.SafeConfigParser()
        config.readfp(config_file)
        settings = dict(config.items('settings'))
        for key, value in settings.items():
            existing_value_type = type(default.get(key, ''))
            if existing_value_type in (list, tuple):
                default[key.lower()] = value.split(",")
            else:
                default[key.lower()] = existing_value_type(value)



from __future__ import absolute_import, division, print_function, unicode_literals

from sys import version_info

import pytest
from frosted import messages as m
from frosted import checker
from pies.overrides import *

from _ast import PyCF_ONLY_AST

from.utils import flakes


def test_undefined():
    flakes('bar', m.UndefinedName)


def test_definedInListComp():
    flakes('[a for a in range(10) if a]')


def test_functionsNeedGlobalScope():
    flakes('''
    class a:
        def b():
            fu
    fu = 1
    ''')


def test_builtins():
    flakes('range(10)')


def test_builtinWindowsError():
    """WindowsError is sometimes a builtin name, so no warning is emitted for using it."""
    flakes('WindowsError')


def test_magicGlobalsFile():
    """Use of the __file magic global should not emit an undefined name
    warning."""
    flakes('__file__')


def test_magicGlobalsBuiltins():
    """Use of the __builtins magic global should not emit an undefined name warning."""
    flakes('__builtins__')


def test_magicGlobalsName():
    """Use of the __name magic global should not emit an undefined name warning."""
    flakes('__name__')


def test_magicGlobalsPath():
    """Use of the __path magic global should not emit an undefined name warning,

    if you refer to it from a file called __init__.py.

    """
    flakes('__path__', m.UndefinedName)
    flakes('__path__', filename='package/__init__.py')


def test_globalImportStar():
    """Can't find undefined names with import *."""
    flakes('from fu import *; bar', m.ImportStarUsed)


def test_localImportStar():
    """A local import * still allows undefined names to be found in upper scopes."""
    flakes('''
    def a():
        from fu import *
    bar
    ''', m.ImportStarUsed, m.UndefinedName)


@pytest.mark.skipif("version_info >= (3,)")
def test_unpackedParameter():
    """Unpacked function parameters create bindings."""
    flakes('''
    def a((bar, baz)):
        bar; baz
    ''')


@pytest.mark.skipif("'todo'")
def test_definedByGlobal():
    """"global" can make an otherwise undefined name in another function defined."""
    flakes('''
    def a(): global fu; fu = 1
    def b(): fu
    ''')


def test_globalInGlobalScope():
    """A global statement in the global scope is ignored."""
    flakes('''
    global x
    def foo():
        print(x)
    ''', m.UndefinedName)


def test_del():
    """Del deletes bindings."""
    flakes('a = 1; del a; a', m.UndefinedName)


def test_delGlobal():
    """Del a global binding from a function."""
    flakes('''
    a = 1
    def f():
        global a
        del a
    a
    ''')


def test_delUndefined():
    """Del an undefined name."""
    flakes('del a', m.UndefinedName)


def test_globalFromNestedScope():
    """Global names are available from nested scopes."""
    flakes('''
    a = 1
    def b():
        def c():
            a
    ''')


def test_laterRedefinedGlobalFromNestedScope():
    """Test that referencing a local name that shadows a global, before it is
    defined, generates a warning."""
    flakes('''
    a = 1
    def fun():
        a
        a = 2
        return a
    ''', m.UndefinedLocal)


def test_laterRedefinedGlobalFromNestedScope2():
    """Test that referencing a local name in a nested scope that shadows a
    global declared in an enclosing scope, before it is defined, generates a
    warning."""
    flakes('''
        a = 1
        def fun():
            global a
            def fun2():
                a
                a = 2
                return a
    ''', m.UndefinedLocal)


def test_intermediateClassScopeIgnored():
    """If a name defined in an enclosing scope is shadowed by a local variable
    and the name is used locally before it is bound, an unbound local warning
    is emitted, even if there is a class scope between the enclosing scope and
    the local scope."""
    flakes('''
    def f():
        x = 1
        class g:
            def h():
                a = x
                x = None
                print(x, a)
        print(x)
    ''', m.UndefinedLocal)


def test_doubleNestingReportsClosestName():
    """Test that referencing a local name in a nested scope that shadows a
    variable declared in two different outer scopes before it is defined in the
    innermost scope generates an UnboundLocal warning which refers to the
    nearest shadowed name."""
    exc = flakes('''
        def a():
            x = 1
            def b():
                x = 2 # line 5
                def c():
                    x
                    x = 3
                    return x
                return x
            return x
    ''', m.UndefinedLocal).messages[0]
    assert 'x' in exc.message
    assert str(5) in exc.message


def test_laterRedefinedGlobalFromNestedScope3():
    """Test that referencing a local name in a nested scope that shadows a
    global, before it is defined, generates a warning."""
    flakes('''
        def fun():
            a = 1
            def fun2():
                a
                a = 1
                return a
            return a
    ''', m.UndefinedLocal)


def test_undefinedAugmentedAssignment():
    flakes(
        '''
        def f(seq):
            a = 0
            seq[a] += 1
            seq[b] /= 2
            c[0] *= 2
            a -= 3
            d += 4
            e[any] = 5
        ''',
        m.UndefinedName,    # b
        m.UndefinedName,    # c
        m.UndefinedName, m.UnusedVariable,  # d
        m.UndefinedName,    # e
    )


def test_nestedClass():
    """Nested classes can access enclosing scope."""
    flakes('''
    def f(foo):
        class C:
            bar = foo
            def f():
                return foo
        return C()

    f(123).f()
    ''')


def test_badNestedClass():
    """Free variables in nested classes must bind at class creation."""
    flakes('''
    def f():
        class C:
            bar = foo
        foo = 456
        return foo
    f()
    ''', m.UndefinedName)


def test_definedAsStarArgs():
    """Star and double-star arg names are defined."""
    flakes('''
    def f(a, *b, **c):
        print(a, b, c)
    ''')


@pytest.mark.skipif("version_info < (3,)")
def test_definedAsStarUnpack():
    """Star names in unpack are defined."""
    flakes('''
    a, *b = range(10)
    print(a, b)
    ''')
    flakes('''
    *a, b = range(10)
    print(a, b)
    ''')
    flakes('''
    a, *b, c = range(10)
    print(a, b, c)
    ''')


@pytest.mark.skipif("version_info < (3,)")
def test_keywordOnlyArgs():
    """Keyword-only arg names are defined."""
    flakes('''
    def f(*, a, b=None):
        print(a, b)
    ''')

    flakes('''
    import default_b
    def f(*, a, b=default_b):
        print(a, b)
    ''')


@pytest.mark.skipif("version_info < (3,)")
def test_keywordOnlyArgsUndefined():
    """Typo in kwonly name."""
    flakes('''
    def f(*, a, b=default_c):
        print(a, b)
    ''', m.UndefinedName)


@pytest.mark.skipif("version_info < (3,)")
def test_annotationUndefined():
    """Undefined annotations."""
    flakes('''
    from abc import note1, note2, note3, note4, note5
    def func(a: note1, *args: note2,
                b: note3=12, **kw: note4) -> note5: pass
    ''')

    flakes('''
    def func():
        d = e = 42
        def func(a: {1, d}) -> (lambda c: e): pass
    ''')


@pytest.mark.skipif("version_info < (3,)")
def test_metaClassUndefined():
    flakes('''
    from abc import ABCMeta
    class A(metaclass=ABCMeta): pass
    ''')


def test_definedInGenExp():
    """Using the loop variable of a generator expression results in no
    warnings."""
    flakes('(a for a in %srange(10) if a)' %
                ('x' if version_info < (3,) else ''))


def test_undefinedWithErrorHandler():
    """Some compatibility code checks explicitly for NameError.

    It should not trigger warnings.

    """
    flakes('''
    try:
        socket_map
    except NameError:
        socket_map = {}
    ''')
    flakes('''
    try:
        _memoryview.contiguous
    except (NameError, AttributeError):
        raise RuntimeError("Python >= 3.3 is required")
    ''')
    # If NameError is not explicitly handled, generate a warning
    flakes('''
    try:
        socket_map
    except Exception:
        socket_map = {}
    ''', m.UndefinedName)
    flakes('''
    try:
        socket_map
    except Exception:
        socket_map = {}
    ''', m.UndefinedName)


def test_definedInClass():
    """Defined name for generator expressions and dict/set comprehension."""
    flakes('''
    class A:
        T = range(10)

        Z = (x for x in T)
        L = [x for x in T]
        B = dict((i, str(i)) for i in T)
    ''')

    if version_info >= (2, 7):
        flakes('''
        class A:
            T = range(10)

            X = {x for x in T}
            Y = {x:x for x in T}
        ''')


def test_impossibleContext():
    """A Name node with an unrecognized context results in a RuntimeError being raised."""
    tree = compile("x = 10", "<test>", "exec", PyCF_ONLY_AST)
    # Make it into something unrecognizable.
    tree.body[0].targets[0].ctx = object()
    with pytest.raises(RuntimeError):
        checker.Checker(tree)



from __future__ import absolute_import, division, print_function, unicode_literals

import textwrap

import pytest
from frosted import messages as m
from pies.overrides import *

from .utils import flakes


def doctestify(input):
    lines = []
    for line in textwrap.dedent(input).splitlines():
        if line.strip() == '':
            pass
        elif (line.startswith(' ') or
                line.startswith('except:') or
                line.startswith('except ') or
                line.startswith('finally:') or
                line.startswith('else:') or
                line.startswith('elif ')):
            line = "... %s" % line
        else:
            line = ">>> %s" % line
        lines.append(line)
    doctestificator = textwrap.dedent('''\
                def doctest_something():
                    """
                        %s
                    """
                ''')
    return doctestificator % "\n       ".join(lines)


def test_doubleNestingReportsClosestName():
    """Lines in doctest are a bit different so we can't use the test from TestUndefinedNames."""
    exc = flakes('''
        def doctest_stuff():
            """
                >>> def a():
                ...     x = 1
                ...     def b():
                ...         x = 2 # line 7 in the file
                ...         def c():
                ...             x
                ...             x = 3
                ...             return x
                ...         return x
                ...     return x

            """
        ''', m.UndefinedLocal).messages[0]

    assert "local variable 'x'" in exc.message and 'line 7' in exc.message


def test_importBeforeDoctest():
    flakes("""
            import foo

            def doctest_stuff():
                '''
                    >>> foo
                '''
            """)


@pytest.mark.skipif("'todo'")
def test_importBeforeAndInDoctest():
    flakes('''
            import foo

            def doctest_stuff():
                """
                    >>> import foo
                    >>> foo
                """

            foo
            ''', m.Redefined)


def test_importInDoctestAndAfter():
    flakes('''
            def doctest_stuff():
                """
                    >>> import foo
                    >>> foo
                """

            import foo
            foo()
            ''')


def test_offsetInDoctests():
    exc = flakes('''

            def doctest_stuff():
                """
                    >>> x # line 5
                """

            ''', m.UndefinedName).messages[0]
    assert exc.lineno == 5
    assert exc.col == 12


def test_offsetInLambdasInDoctests():
    exc = flakes('''

            def doctest_stuff():
                """
                    >>> lambda: x # line 5
                """

            ''', m.UndefinedName).messages[0]
    assert exc.lineno == 5
    assert exc.col == 20


def test_offsetAfterDoctests():
    exc = flakes('''

            def doctest_stuff():
                """
                    >>> x = 5
                """

            x

            ''', m.UndefinedName).messages[0]
    assert exc.lineno == 8
    assert exc.col == 0


def test_syntax_errorInDoctest():
    exceptions = flakes(
            '''
            def doctest_stuff():
                """
                    >>> from # line 4
                    >>> fortytwo = 42
                    >>> except Exception:
                """
            ''',
            m.DoctestSyntaxError,
            m.DoctestSyntaxError).messages
    exc = exceptions[0]
    assert exc.lineno == 4
    assert exc.col == 26
    exc = exceptions[1]
    assert exc.lineno == 6
    assert exc.col == 18


def test_indentationErrorInDoctest():
    exc = flakes('''
            def doctest_stuff():
                """
                    >>> if True:
                    ... pass
                """
            ''', m.DoctestSyntaxError).messages[0]
    assert exc.lineno == 5
    assert exc.col == 16


def test_offsetWithMultiLineArgs():
    (exc1, exc2) = flakes(
            '''
            def doctest_stuff(arg1,
                                arg2,
                                arg3):
                """
                    >>> assert
                    >>> this
                """
            ''',
        m.DoctestSyntaxError,
        m.UndefinedName).messages
    assert exc1.lineno == 6
    assert exc1.col == 19
    assert exc2.lineno == 7
    assert exc2.col == 12


def test_doctestCanReferToFunction():
    flakes("""
            def foo():
                '''
                    >>> foo
                '''
            """)


def test_doctestCanReferToClass():
    flakes("""
            class Foo():
                '''
                    >>> Foo
                '''
                def bar(self):
                    '''
                        >>> Foo
                    '''
            """)


def test_noOffsetSyntaxErrorInDoctest():
    exceptions = flakes('''
            def buildurl(base, *args, **kwargs):
                """
                >>> buildurl('/blah.php', ('a', '&'), ('b', '=')
                '/blah.php?a=%26&b=%3D'
                >>> buildurl('/blah.php', a='&', 'b'='=')
                '/blah.php?b=%3D&a=%26'
                """
                pass
            ''',
        m.DoctestSyntaxError,
        m.DoctestSyntaxError).messages
    exc = exceptions[0]
    assert exc.lineno == 4
    exc = exceptions[1]
    assert exc.lineno == 6
